/*
  h_file.c   Nov98-Jan17  JoergS  GPL
  
  this is a file for handling Htmp-File (tmp_hxy.000) 
  reasons and goals for this file:
  - 2GB file breaking => use LFS (64bit) or more blocks
  - error handling (perror nach printf???) Jun99
  - open/close
  - hidden variables (encapsulating of file-handling)
  - better reading of main-file
  Ideas:
  - chksum for tmp-file, check during read?
  - vvv==1 ? measure disk performance
  - compressed tmp_hxy.XXX ? (example: 1GB => 336MB, zip=17m  unzip=3m read=1m)
                        thor  gzip -c1 1GB => 640MB, zip=10m  unzip=3m
                        thor  gzip -c9 1GB => 579MB, zip=125m unzip=3m 
                        thor  gzip -c4 1GB => 550MB, zip=14m  unzip=3m
  - see pipe.c!!! (h_file,fopen,fclose,... & fn,pipe_ropen,... kapseln
  - repeat H*v if error!?
  - see also utils/vbuf.c for buffering/speed/IO
  
  this file is still changing Nov98,Apr99,Jun99,Nov99
  history:
   v2.17 - array of thxy -to- thxy of arrays
   v2.0.1 - access to scplx only via Macros M_d2f(), M_f2d()
   v2.0.0 - file-split removed (see history)
   03.02.00 - bug cplx rr => scplx rr
   25.11.99 - bug in my_fread/write if file > 1MB beseitigt
   07.01.00 - bug auf MIPS beseitigt (high-low Format!!!)
               Vorsicht bei type-Umwandlungen (high-low/low-high format)
   ToDo: remove pipe.c (simpler), add mem-only-version (mmap?) /dev/shm???

*/
#include <stdio.h>
#include <stdlib.h>
/* #include <errno.h> 2015-12-07 not needed? t100.gpu02 no errno */
#include <string.h>	/* strerror */
#include "defs.h"
#include "spins.h"
#include "h_file.h"
#include "pipe.h"
#include "block.h"
#if 1
#include <unistd.h>		/* test this with configure! */
#define SLEEP(x) sleep(x)
#else
#define SLEEP(x) {}
#endif

#ifndef TMPDIR        /* 2018-11 added for more flexibility */
#define TMPDIR "tmp/"  /* alternatives: "./" "../tmp/" "/tmp/" */
#endif

/*
 * v2.30 can have different numbers of I/O threads and compute threads
 */

#ifndef HRMAX
#if HRMAX>0
#error "HRMAX!=0 if not defined?" 
#endif
#endif
#if HRMAX>0	/* table for 16bit index to Hamilton-entries */
 int  hr_len[B_NUM];    /* number of table entries */
 cplx hr[B_NUM][HRMAX];	/* table of matrix values, one table per thread */
 /* better use realloc for dyn. pt_n ? */
#endif

/* define the basename for the H-files htmpBBBB.PPP
 * where BBB is the IO-thread-number (block)
 * and PPP the paramter index (ToDo)
 */
#define Htmp "htmp"	/* ./tmp can be a link to /tmp/$USER */

/* public variables */

/* privat variables */
char  h_name[B_NUM][40]; /* file-name B_NUM=(ToDo: <=)B_NUM files */
int   h_error=0;         /* set if H could not be written */
int   h_state[B_NUM];    /* 0=closed, 1=ropen, 2=wopen */
int   h_open[B_NUM];    /* number cthreads have files opened 0..B_NUM/F_NUM-1 */
static int dbg_wcnt=0, dbg_rcnt=0;   /* number of debug outputs (write/read) */
/* 2013-04: allow less but interleaved block files on shared slow disks  
 *   disk speed may vary between 100% to less than 10% for parallel files
 *  but each task must send at least zero block until it gets ready_for_close
 *  ToDo: alternatively synchronize blockwrites using ring-token
 *        num_IO_synced_threads instead of threads per file
 *  ring_tokens bit0=ORed_some_thrads_had_IO, bit1=1(no_IOs_last_round=close)
 */
/* int threads_per_file=1; */ /* default: 1 = 1 file/task, pt_n = 1 file/node */

/* my comfortable erroroutput for misterious errors */
int my_ferror(const char *msg){ /* JS-Mar11 add const */
  perror(msg);	/* errno_msg => stderr */
  SLEEP(1);
  return 0;
}

/* including error-handling (public???) */
size_t my_fread( char *ptr, size_t size, size_t nmemb, int block){
  size_t n_readed=0,nf,ni,tries=0; /* sgi-CC can not add void* */
  nmemb*=size;
  if (size==0) return nmemb; /* 2017-09-13 n1=4 -t3 0div */
  if (block>=B_NUM) ERR(9999);
  if ( nmemb && ptr)
  while(n_readed< size /* nmemb */ ){ /* mind. 1 element */
    /* some clibs do alloc a buffer for disk_read which cost time if it is to big */
    ni=nmemb-n_readed;
    /* v2.29 add tries>0 because I forgot whats the splitting is good for */
    if(tries>0 && ni>262144) ni=262144;	/* 2^n=256kB pieces, why? */
    nf=pipe_read(ptr,(int)ni,block);
    if(nf<=0){ E3("read=%d ni=%d block=%d ",(int)nf,(int)ni,block);
               my_ferror("read");if(tries++<5)continue;break; }
    n_readed+=nf; ptr+=nf; tries=0;
  } /* while */
  if (mpi_i*pt_n+block<=8 && ((vvv&16)||!ptr) && dbg_rcnt<64) { dbg_rcnt++;
    if (dbg_rcnt==1) /* minimize output differing from run to run */
      printf("DBG.fread  %2d.%2d p= %p\n", mpi_i, block, ptr);
    printf("DBG.fread  %2d.%d i=%dK o=%dK err=%d cnt=%5d\n",
           mpi_i, block, (int)(nmemb/1024),
           (int)(n_readed/1024), h_error, dbg_rcnt); }
  return n_readed/size;
}

int h_wopen(int block); /* declaration, defined below */
int h_ropen(int block); /* declaration, defined below */
/* 
   write large objects in small 1e6-pieces,
    because fwrite (probably) need lot of memory
   size should be Hsize = multiple of sector size ! (4k and bigger)
   ToDo17: DBG only on log-iteration
   ToDo18: open is slow on HPC-cluster, open on first read (no err) here
 */
size_t my_fwrite( void *ptr, size_t size, size_t nmemb, int block){
  size_t n_written=0, nf, ni, tries=0;
  char *p=(char *)ptr;	/* sgi-CC can not add void* */
  // if (h_state[block]!=2) ERR(9999); /* state != not wopen */
  /* if (h_state[block]!=2) file not wopen # 0=close,1=ropen,2=wopen */
 //   do not close open file, may be this hides potential bugs later
 // if (h_state[block]==1) { /* still ropen, but no error */
 //   if (h_rclose(block)) h_state[block]|=16; /* rclose error */
 // }
  if (h_state[block]==0) { /* still closed, but no error, 2018-11-11 */
    if (h_wopen(block))  h_state[block]|=8; /* wopen error */
  }
  nmemb *= size;
  if (ptr && (!h_error) && h_state[block]==2)
  while (n_written<nmemb) { ni= nmemb - n_written;
#if STOREH >= 2 && STOREH <= 4   /* dont split malloc blocks, JS0904 */
    if (tries>0 && ni>262144) ni=262144; /* 256kB pieces, avoiding OOM? */
#endif
    nf=pipe_write(p,(int)ni,block); // return int?
    if (nf!=(size_t)-1) { n_written+=nf; p+=nf; }
    if(nf!=ni){ ERR(620); if (h_error) break;  /* dont try 3x on old errors */
      /* may happen on O_DIRECT and blocksize512 != filepos */
      tries++; sleep(1<<(2*tries)); /* time to relax 1,4,16,64s */
      if (tries<3) continue; /* try 3 times */
      h_error=1; break;    /* write finally failed */
    } tries=0; /* reset cnt */
  } /* end while */
  if (mpi_i*pt_n+block<=8 && ((vvv&16)||!ptr) && dbg_wcnt<64) { dbg_wcnt++;
    if (dbg_wcnt==1) /* minimize output differing from run to run */
      printf("DBG.fwrite %2d.%2d p= %p\n", mpi_i, block, ptr);
    printf("DBG.fwrite %2d.%2d i=%dK o=%dK err=%d cnt=%5d\n",
           mpi_i, block, (int)(nmemb/1024), 
           (int)(n_written/1024), h_error, dbg_wcnt); }
  return n_written/size;
}

/* --- end of file handler --- */

/* --------- interface between compute blocks and H storage ------ */
/* open for reading */
int h_ropen(int block){ /* 0..B_NUM-1, block > B_NUM possible */
  static int nout=0;
  if (mpi_i*pt_n+block<=8 && (vvv&16) && ++nout<4*pt_n) /* max 8 Lines debugging or no output */
    printf("DBG.h_ropen  %2d.%d %s %d\n", mpi_i, block,
            h_name[block], h_open[block]);
  if (block<0 || block>=pt_n) ERR(9999); /* for safety */
  if (++h_open[block%B_NUM]==1) { /* open stream for first block */
    if ( pipe_ropen(h_name[block],block) ){
      E1("fopen_%s",h_name[block]);return 1; }
  }
  b_bnz[block]=HBLen; /* force h_read on first h_get */
  if (h_state[block]!=0) { F1("h_ropen(%d): file was open", block); }
  h_state[block]=1;  /* opened for read */
  return 0;
}

/* open for writing */
int h_wopen(int block){  /* 0..B_NUM-1, block > B_NUM possible */
  if (block<0 || block>=pt_n) ERR(9999); /* for safety */
  /* JS-Apr2011 for better diff output, 0000...%04d mpi_n*pt_n-1 replaced */
  if(vvv && mpi_i*pt_n+block==0)
    printf(" write/store " TMPDIR Htmp "XXXX STOREH=%d(1=gz,2=io,3=bz,4=buf_io,...)",
           STOREH);  /* Hsize = size.struct...[HBlen] */
  if ((vvv&16) && mpi_i*pt_n+block<8) usleep((mpi_i*pt_n+block+1)*10000);
  if ((vvv&16) && mpi_i*pt_n+block<8) /* maximum debug info for 8 threads */
     printf("\nDBG.h_wopen %2d.%d open=%d",
             mpi_i, block, h_open[block%B_NUM]);
  if (++h_open[block%B_NUM]==1) { /* open stream for first block */
    if ((vvv&16) && mpi_i*pt_n+block<8)
       printf(" pipe_wopen_%d.%d", mpi_i, block);
    if ( pipe_wopen(h_name[block%B_NUM],block%B_NUM) ) { 
      E1("wopen %s",h_name[block%B_NUM]); h_error=1; return 1; }
  }
  b_bnz[block]=0;
  b_bn1[block]=0; if (h_state[block]!=0) ERR(9999); /* file not open */
  h_state[block]=2;  /* opened for write */
  if ((vvv&16) && mpi_i*pt_n+block<8) printf(" ok");
  if ((vvv&16) && mpi_i*pt_n+block<8) usleep((8-mpi_i*pt_n+block)*10000);
  return 0;
}

int h_rclose(int block) {  static int nout=0;
  if (mpi_i*pt_n+block<8 && (vvv&16) && ++nout<4*pt_n) /* max 8 lines output */
    printf("\nDBG.h_rclose %2d.%d open=%d", mpi_i, block, h_open[block%B_NUM]);
  // if (h_state[block]!=1) ERR(9999); /* file not open */
  h_state[block]=0;                 /* mark as closed */
  if (--h_open[block%B_NUM]==0) /* close stream if last block flushed */
    pipe_close(block%B_NUM); /* ToDo return errors, state+16 on close-error */
  return 0;
}

/* --- read from Htmp-file, ret: 0=ok --- 
 * read from disk into H-buffer, struct thxy = xy[HBLen], r[HBLen](or idx+para)
 *  ToDo: get special block from stream of random(?) stored blocks 
 *  search dynamic buffer, if not found extend buffer until block found
 *  max. B_NUM/B_NUM (* PARAM)?
 * this buffer is written by B_NUM IO-threads
 *   and read by B_NUM compute threads (v2.26 deadlocks?)
 * IO-thread=file blocks=compute-threads
 *                          v-blocks IO=0-Buffer? ToDo
 *    0:          0,1,4,0,5,0,1,4 (compute block 0 was faster than the others?)
 *    1:          2,3,6,7,2,3,6,7
 */
 
/* privat */
/* read special block, !!! ToDo: read next block, caller must handle that
 *   read from stream[blk] or memory[blk] to tbh (i.e. b_hbuf[block])
 *   for STOREH==0 we simply return tbh as pointer to next H_block 
 * 2017-02-20 add parameter int len, 2017-04-06 size_t len (-m32 failed)
 */
struct thxy *h_read(struct thxy *tbh, size_t len, int block){
  if (((vvv&16) || !tbh) && dbg_rcnt==0)  /* minimize diff output per run */
    printf("DBG.h_read %2d.%d p= %10p\n", mpi_i, block, (void *)tbh);
  if (((vvv&16) || !tbh) && dbg_rcnt<64)  /* print before, because read may hang */
    printf("DBG.h_read %2d.%d state=%d len= %6lu\n", mpi_i, block, 
           h_state[block], (ULong)len );
  if (len==0 || !tbh) return tbh; /* 2017-01. 2017-09-13 n=4 -t3 */
  /* ToDo18: open is slow on HPC-cluster, open on first read (no err) here */
  if (h_state[block]==0) { /* still closed, but no error, 2018-11-11 */
    if (h_ropen(block))  h_state[block]|=16; /* ropen error */
  }
  if (h_state[block]!=1) ERR(9999); /* file not open */
  if (my_fread((char *)tbh/*&b_hbuf[block]*/,len,1,block)!=1){
     E0("h_read"); ERR(9622); return NULL;
   }
  if ((vvv&16) && tbh && dbg_rcnt<100) { dbg_rcnt++;  /* move to h_read */
    printf("DBG.h_read %2d.%d state=%d  j %6ld\n",
          mpi_i, block, h_state[block],
          ((tbh)?((long)tbh->nz[ 0 ].xy&~XY_FLAG):0l)); }
  return tbh;  /* return same buffer, data copied in */
}


/* ------------------------------------------------------------------ *
 * h_put: write matrix-element and its column (line encoded)          *
 * h_get: read  matrix-element and its column (line+1 if column>n1)   *
 * ------------------------------------------------------------------ */
#define HR_SIZE   sizeof(HR_TYPE)
#define XY_SIZE   sizeof(XY_TYPE)  


#if HRMAX>0
/* seperate error routine, only loaded to CPU-cache in error-case */
void err_fulltable(){
    int i, j;
    /* output maximum on information, what was went wrong? */
    printf("\nerr_fulltable: --debug--\n");
    for (i=0;i<pt_n;i++) if (hr_len[i]>=HRMAX-4) {
      printf(" Hblock=[%d.%d] b_hnz=%9ld hr_len=%4d hr[]=",
             mpi_i, i, b_hnz[i], hr_len[i]);
      for (j=0;j<hr_len[i] && j<100;j++) printf(" %g",D(hr[i][j]));
      if (j>=100) printf(" ...");
      printf("\n");
    }
    ERR(9100);
}
#endif

/* h_put fills a write buffer, this fkt writes the buffer out!
 * h_wclose is only called to close the file after writing
 * block=1..B_NUM-1 can be equal and greater than F_NUM to flush all buffers
 * called by storeh for every thread + from storeh2 thread0
 */
int h_wclose(int block){  /* store in file */
  /* JS0904: returning on error causes more errors if H partly stored */
  /* if(h_error) return 1; */ /* bad, if partly stored */
  if (vvv&16)
    printf("\nDBG.h_wclose %2d.%d open=%d", mpi_i, block, h_open[block]);
  // if (h_state[block]!=2) ERR(9999); /* file not open */
  h_state[block]=0; /* mark as closed */
  if (--h_open[block]==0) { /* close stream if last block flushed */
    if (vvv&16) printf(" pipe_close %2d.%d", mpi_i, block);
    pipe_close(block); }
  return 0;
}

/* h_put and h_get moved to spins.c
 * to make the call shorter and allow inlining => no speedup Jan07 */

/* initialize H-storage, called by spins.c storeh*(), thread0 only */
void ini_thxy(ULong n1){ /* alloc */
//  static int num_ini=0;   /* malloc only once */
  int i1,i2,j; ULong i; XY_TYPE xy;
  // static struct thxy t1; /* 2018-11-11 fat + unused (not on stack) */
//  if (!num_ini) {
     // if (!b_hbuf) b_hbuf=(struct thxy *)malloc(pt_n*sizeof(struct thxy));
     // if (!b_hbuf) ERR(9999);
//  } num_ini++;
// printf("L%d n1=%f\n",__LINE__,n1*1.);
  for(j=0;j<pt_n;j++){
    h_state[j]=0; /* default 0=close */
#if HRMAX>0
    hr_len[j]=0; /* table for index-to-matrix value */
#endif
  }
  for(i1=0;i1<B_NUM;i1++){
    h_open[i1]=0;               /* no bocks open */
    h_name[i1][0]=0;		/* file-name ToDo: set name here! */
    sprintf(h_name[i1],TMPDIR Htmp "%04u.%03u",(int)i1+mpi_i*pt_n,1);
  }
  /* measure: i1 = used bytes for xy + XY_FLAG,
   *          i2 = sizeof xy in bytes (ToDo: check sense)
   * we store 0..node_n1 (not n1) to .xy, less bits
   * i must be unsigned to avoid endless loops if first bit set
   */
  for (j =0,i=2*     n1;i!=0; j++, i>>=1); /* n1 bits + FLAG (not needed) */
  for (i1=0,i=2*node_n1;i!=0;i1++, i>>=1); /* node_n1 bits + FLAG (Jan09) */
  for (i2=0,xy=1;      xy!=0;i2++,xy<<=1);
  if (vvv && !mpi_i) 
    printf("\n ini_thxy.bits:  thxy.xy %d  BY %d  HR %d  node_n1,n1 %d %d",
       i2,  8*(int)sizeof(BY_TYPE), 8*(int)sizeof(HR_TYPE), i1, j);
  if (  i1> i2) F0("xy to small, enlarge h_file.h:XY_TYPE or add mpi-tasks");
  if (2*i1<=i2 && i1>20 && !mpi_i) /* dont warn on play systems */
    printf("\nWarning: reduce h_file.h:XY_TYPE to spare memory.");
  /* h_file.h struct thxy XY_TYPE */
  j=sizeof(int); /* avoid compiler warnings for if(const1==const2) */
  if (j!=sizeof(XY_TYPE) && !mpi_i) E0("sizeof int!=thxz.xy"); /* ToDo: why? */
  if (vvv && !mpi_i) printf("\n"); /* multithread nicer ...\n */
}

#define HR_FILE "hr_s%04d.dat"  /* ToDo: mpi.001 etc. */

/* checkpoint Hstorage, write restore data to HR_FILE
 ToDo: writing to mpi_n local files(?) (distributed or shared medium)
 ToDo: store a unique hash to validate having correct files
   get_max_hr MPI_IO_open + seek(mpi_i*max_hr*hrsize) + save (shared file?)
   suspend/resume only if mpi_n not changed !? 
*/
extern double opt_maxfile; /* spins.c */
void h_checkpoint(int vvv){
  char s1[80]="dummy_checkpoint";
  int i, hrmax;
  FILE *f1=NULL;
  /* if( save_mode & 1 ) --- small amount of data! save every time */
  /* save data before possible crash */
  sprintf(s1,TMPDIR HR_FILE,mpi_i);
  if (opt_maxfile!=0 && mpi_n*pt_n<=4) /* bad for thousend of files */
  { f1=fopen(s1,"wb"); if (!f1) E1("error fopen %s",s1); } /* mpi*pt */

  hrmax=0; /* max hr-table_length */
  for (i=0;i<pt_n;i++) {  /* all pthreads */
#if HRMAX>0
    if (hr_len[i]>hrmax) hrmax=hr_len[i];
    /* ToDo: better output num threads*mpi + hr_min/max */
    // ToDo: MPI_.... min,max output? loop !?
    if ((!mpi_i) && i==0)
      printf("\nDBG: h_checkpoint.w thread=0..%d ...\n", pt_n);
    if (mpi_i==1 && i==0) /* add i==0 2017-01-11 */
      puts("\nWARNING: h_checkpoint not ready for MPI code");
    if (vvv && mpi_i==0 && i<=8)
      printf(((i==0)?" hr[] len: %d":" %d"),hr_len[i]);
    if (f1){
      if (fwrite(&hr_len[i],sizeof(hr_len[i]), 1, f1)!=1
       || fwrite(&hr[i][0], sizeof(hr[i][0]), hr_len[i], f1)
                                    !=(size_t)hr_len[i]) 
        E1("h_checkpoint.fwrite pt_i=%d", i); /* hr_len==0 ??? */
    }
#endif
    if (f1) /* 2012-01 ToDo: within pt-loop?, 2012-04 add error handling */
      if (fwrite(b_hnz, sizeof(b_hnz[0]), pt_n, f1)!=(size_t)pt_n
       || fwrite(b_snz, sizeof(b_snz[0]), pt_n, f1)!=(size_t)pt_n
      /* || fwrite(b_sn1, sizeof(b_sn1[0]), pt_n, f1)!=(size_t)pt_n*/)
        E0("h_checkkpoint");
  } // i= all pthreads
#ifdef CONFIG_MPI
  MPI_Allreduce(&hrmax,&i,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD); hrmax=i;
#endif
  if (f1) fclose(f1);
  if (vvv && !mpi_i) { /* to minimize hr_bits in H, we have to know hr_num */
    printf("\nDBG: ToDo better hr[param], extra hr_sqrt(n2[i]), hr_phase?");
    printf("\nDBG: h_checkpoint.w: hr_num_max= %5d (entries/thread) %s..\n",
           hrmax, s1);
  }
}

/* restore H-storage, after h_checkpoint, same nodes (local disks) */
/* load data after crash, defekt node must be replaced? better recalculate! */
void h_restore(int vvv){
  int hrmax=0;
  FILE *f1=NULL;
  if (!mpi_i) fputs(" h_restore(chkpt>=4,STOREH!=0 only) ...",stdout);
#if STOREH != 0  /* makes no sence for H in memory, v2.32 */
  char s1[80];
  sprintf(s1,TMPDIR HR_FILE,mpi_i);
  f1=fopen(s1,"rb");if(!f1)E0("error fopen");
#endif
#if HRMAX>0
  if (f1) { int i;
  for(hrmax=i=0;i<pt_n;i++){
    if (vvv) printf(" load hr[%d]=",i);
    if (fread(&hr_len[i],sizeof(hr_len[i]),1,f1)!=1  // ToDo: correct? 2012
    ||  fread(&hr[i][0], sizeof(hr[i][0]), hr_len[i],f1)!=(size_t)hr_len[i])
      F0("h_restore.fread.HR");
    if ( hr_len[i]>hrmax ) hrmax=hr_len[i];
    if (vvv) printf("%d",hr_len[i]);
  } } // if (f1)
#endif
  if (f1)
    if (fread(b_hnz, sizeof(b_hnz[0]), pt_n, f1)!=(size_t)pt_n
    ||  fread(b_snz, sizeof(b_snz[0]), pt_n, f1)!=(size_t)pt_n
  /*  ||  fread(b_sn1, sizeof(b_sn1[0]), pt_n, f1)!=(size_t)pt_n */)
     F0("h_restore.fread");
  if (f1) fclose(f1);
#ifdef CONFIG_MPI /* ToDo: bad code !? */
  { int i;
    MPI_Allreduce(&hrmax,&i,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD); hrmax=i; }
#endif
  if (vvv && !mpi_i)
    printf(" h_restore: max hr_num= %5d (entries/block)\n",hrmax);
}
/* ---------------------------------------------------------- */
