#include <string.h>
#include <stdlib.h> /* added 2016-04-21 */
#include <stdio.h>  /* added 2016-04-21 */
#include "defs.h"
#include "block.h"
#include "array_m.h"
#include "vector.h"
#include "hilbert.h"
#include "spins.h"
//#ifdef CONFIG_MPI
//#include <mpi.h> /* already in block.h */
//#endif

struct vblock {     /* arguments for working threads */
  int   blk;        /* block (thread) nr */
  scplx *v0;        /* vector (used by clrvec,mltply,addav) */
  double r;         /* factor (used by mltply,addav) */
  scplx *v1;        /* vector (used by addav) */
  /* ToDo17 add srand per thread + thread local random */
};

/* void-function for pthreads, v1[]=a*v0[] */
void copy_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  scplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  scplx *v1=((struct vblock *)ptr)->v1+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r, n2t=0;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) {
    v1[i]  = M_d2f(M_f2d(v0[i])*r);
    /* n2t   += Norm(M_f2d(v1[i])); */
  }
  ((struct vblock *)ptr)->r = n2t; /* scalar product (v1,v1) */
  return;
}

double copy_av(scplx *v1, double r, scplx *v0){  /* v1=r*v0, [return norm(v1)] v2.27 */
  struct vblock buf[B_NUM];
  int blk; double n2=0;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    buf[blk].v1 =v1;
    parallel_start( blk, copy_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) {
    parallel_wait(blk); /* synchronize */
    n2 += buf[blk].r;
  }
#if 0
#ifdef CONFIG_MPI
  { double add=n2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    n2=sum; }
#endif
#endif
  return(n2);
}

/* void-function for pthreads */
void add_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  scplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  scplx *v1=((struct vblock *)ptr)->v1+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r, n2t=0;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) {
    /* 2013-04-04 avoid float16 += float16 */
    v1[i] =  M_d2f(M_f2d(v1[i]) + M_f2d(v0[i])*r);
    n2t   +=  Norm(M_f2d(v1[i]));
  }
  ((struct vblock *)ptr)->r = n2t; /* scalar product (v1,v1) */
  return;
}

double add_av(scplx *v1, double r, scplx *v0){  /* v1+=r*v0, return norm(v1) */
  struct vblock buf[B_NUM];
  int blk; double n2=0;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    buf[blk].v1 =v1;
    parallel_start( blk, add_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) {
    parallel_wait(blk); /* synchronize */
    n2 += buf[blk].r;
  }
#ifdef CONFIG_MPI
  { double add=n2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    n2=sum; }
#endif
  return(n2);
}

/* void-function for pthreads */
void mltply_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  scplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) 
    v0[i] = M_d2f(M_f2d(v0[i])*r);
  return;
}

void mltply_av(double r, scplx *v0){  /* v0*=r */
  struct vblock buf[B_NUM];
  int blk;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    parallel_start( blk, mltply_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) parallel_wait(blk); /* synchronize */
  return;
}

/* void-function for pthreads */
void clrvec_block( void *ptr ){
  const int blk=((struct vblock *)ptr)->blk;
  scplx      *v=((struct vblock *)ptr)->v0+b_ofs[blk];
  ULong i;
  for (i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) v[i]=M_d2f(0);
  return;
}

void clrvec(scplx *v){ int blk;
  /* its important to inititialize memory locally in the threads for NUMA */
  struct vblock buf[B_NUM];
  /* if (vvv && !mpi_i) printf("\n clrvec"); */
  // if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  // JS-2012-07 add error check (MPI.sum_ERR?), 2012-10 case mpi_n > n1
  if ((!v) && b_ofs[pt_n]) F0("!v"); 
  for (blk=0;blk<pt_n;blk++) {
#if 0
    if (vvv&4 && (mpi_i*pt_n+blk<4 || mpi_i*pt_n+blk>=mpi_n*pt_n-4))
      printf("DBG: v_%d_%d[%lu]=0\n", mpi_i, blk, b_ofs[blk+1]-b_ofs[blk]);
#endif
    buf[blk].blk=blk;
    buf[blk].v0 =v;
    parallel_start( blk, clrvec_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) parallel_wait(blk); /* synchronize */
  // if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  /* if (vvv && !mpi_i) printf(" ok"); */
  return;
}

/* fast access to 4K-memory pages, assign mempages to CPU cores 2015-03-09 */
/* void-function for pthreads */
void touchvec_block( void *ptr ){
  int  blk=((struct vblock *)ptr)->blk;
  scplx *v=((struct vblock *)ptr)->v0+b_ofs[blk];
  ULong i;
  for (i=0l;i<b_ofs[blk+1]-b_ofs[blk];i+=4096) v[i]=M_d2f(0);
  return;
}

void touchvec(scplx *v){ int blk;
  /* its important to inititialize memory locally in the threads for NUMA */
  struct vblock buf[B_NUM];
  if ((vvv&4) && !mpi_i) printf(" touchvec");
  if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  // JS-2012-07 add error check (MPI.sum_ERR?), 2012-10 case mpi_n > n1
  if ((!v) && b_ofs[pt_n]) F0("!v"); 
  for (blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].v0 =v;
    parallel_start( blk, touchvec_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) parallel_wait(blk); /* synchronize */
  if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  if ((vvv&4) && !mpi_i) printf(" ok\n");
  return;
}

double normvec(scplx *v){ /* non-thread-parallel, ToDo: norm_v() */
  double r=0.0; ULong j;
  for (j=0l;j<b_ofs[pt_n];j++)  r += Norm(M_f2d(v[j]));
#ifdef CONFIG_MPI
  { double add=r, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=sum; }
#endif
  if (r<=1e-20){ if(!mpi_i) printf(" normvec.r=%.0e",r); } else
  if (r!=1.0){ 
    if (!mpi_i) printf(" normvec.r=1%+.0e",r-1);
    r=1.0/sqrt(r); 
    /* if (vvv&1 && !mpi_i) printf(" 1-r*(1/r)=%.0e",1-r*(1/r)); */
    /* if (!mpi_i) printf(" normalize"); */
    if (!mpi_i) printf(" v/r"); /* shorter */
    mltply_av(r,v); // v=r*v;
    for (r=j=0l;j<b_ofs[pt_n];j++) r += Norm(M_f2d(v[j]));
#ifdef CONFIG_MPI
  { double add=r, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=sum; }
#endif
    if (vvv&1 && !mpi_i) printf(" r=1%+.0e",r-1);
  }
  return r;
}

/* if blk==-1 save all blocks, idx>=0 index for name="d%d.vec" */
/* called only by thread blk==0 
 * MPI: writing into one file by (open+append+close+nexttask) does not work
 *   test 2012-05-15 mkautz.nfs, so we use MPI_IO (bad data, extreme slow)
 */
int savevec(scplx *v, const char *nam, int idx, int blk){
#if CONFIG_MPI
  // MPI_Offset my_offset; // , my_current_offset;   // offset = integer?
  // MPI_File fh;
  MPI_Status status;
#endif
  int i=0, nerr=0; char s1[128], *s2;
  if (blk>0) {
    printf("\nWarning savevec: code called from working thread, skip");
    return 0;
  }
  if (idx>=0) i=sprintf(s2=s1,nam,idx); else s2=(char *)nam;
  if (i<0) E0("fname to long");
  if ((vvv&5)==1 && !mpi_i) printf(" wv");      // short form
  if ((vvv&4)==4 && !mpi_i) printf(" w.%s",s2); // long form
#if CONFIG_MPI  && 0 // 2012-06-06 disabled fails on q.NFS
  // 2012-05-15 (120qu.nfs3.ADIOI_Set_lock failed+exit, needs noac)
  //  File locking failed in ADIOI_Set_lock(fd 6,cmd F_SETLKW/7,type F_WRLCK/1,whence 0) with return value FFFFFFFF and errno 25.
  //  If the file system is NFS, you need to use NFS version 3, ensure that the lockd daemon is running on all the machines, and mount the directory with the 'noac' option (no attribute caching).
  // ToDo: transfer in 2MB chunks to task0 (io-task)
  MPI_File_open(MPI_COMM_WORLD, s2, MPI_MODE_CREATE
                                  | MPI_MODE_WRONLY, MPI_INFO_NULL,&fh);
  my_offset= mpi_ofs[mpi_i]*Csize;
  MPI_File_seek(fh, my_offset, MPI_SEEK_SET);
  // MPI_File_get_position(fh, &my_current_offset);
  if (((int)b_ofs[pt_n])*Csize!=b_ofs[pt_n]*1.*Csize) F0("int_overflow");
  MPI_File_write(fh, (void *)v, b_ofs[pt_n]*Csize, MPI_CHAR, &status);
  // MPI_Get_count(&status, MPI_CHAR, &count); // ToDo: error checking
  MPI_File_close(&fh);
#else // 0
  // ToDo: send to task0, task0 recvs from all tasks and writes (2MB chunks)
  // 2012-06-07 works for NFS without noac
  if (!mpi_i) {
    FILE *f=NULL; ULong n2;
    if (!(f=fopen(s2,"wb"))) { E1("fopen(%s)",s2);perror("");nerr=1; }
    // old parallel: v==0 happens for tasks where node_n1==0 (i.A. n1 < mpi_n)
    if ((!nerr) && b_ofs[pt_n] && v) {
      n2=fwrite((void *)v,Csize,(size_t)b_ofs[pt_n],f); /* full node */
      if (n2!=b_ofs[pt_n]){E2("write=%lu err=%d",n2,ferror(f));nerr=2;}
    }
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
    if (mpi_n>1) { // task0 recv from all other tasks
      int i1, i3, n3; ULong i2; char *buf;
      /* 4MB chunks transfered and saved for max performance (old 16MB) 2015-04 
       *   must be a compromize of speed and memory usage (OOM) */
#define VChunkSize (4<<20)  /* 4MB */
      buf=(char *)malloc(VChunkSize); // 16MB  2015-02: fails if mem is filled by H
      if (!buf) { E1("malloc%dMB", VChunkSize>>20); nerr=-2; } 
      // ToDo: reduce chunksize on err, send bsize?
      for (i1=1; i1<mpi_n; i1++) {
        int b2= (buf)?0:-1; // 0 = ok, -1 = error abort
        // printf("DBG.t0: send token %d to i1=%d\n",b2,i1);
        MPI_Send(&b2,1,MPI_INT,i1,1,MPI_COMM_WORLD); // send token to i1
        if (buf)
        for (i2=mpi_ofs[i1]*Csize; i2<mpi_ofs[i1+1]*Csize; i2+=VChunkSize) {
          if (i2+(VChunkSize)<=mpi_ofs[i1+1]*Csize) i3=VChunkSize;
          else              i3=mpi_ofs[i1+1]*Csize-i2;
          MPI_Recv(buf,i3,MPI_BYTE,/*src*/i1,/*tag*/0,MPI_COMM_WORLD,&status);
          // MPI_Get_count(&status, MPI_BYTE, &i3 /* out.count */);
          if (f) n3=fwrite(buf,1,i3,f); else n3=0; /* full node */
          if (f && n3!=i3){ E2("write=%d err=%d",i3,ferror(f));nerr=2; }
        } // loop over transfer chunks
      } // loop over tasks
      if (buf) free(buf); // libc may still not free for later use
    } // mpi_n>1
#endif
    if (f) fclose(f);  // ToDo: fsync (chkpt!)
  } // mpi_i==0
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
  if (mpi_i!=0) { // mpi_i!=0   send to task0
    ULong i2; int i3, b2;
    //  n1=80 = 0,16,32,48,64 + 80(rest0) (16 Byte chunks)
    //  n1=77 = 0,16,32,48    + 64(rest13)
    // 2012-11-14 JS problems with mpich, without token ...
    //    ... all nodes send data and n0.mpich may buffer and crash on OOM
    MPI_Recv(&b2,1,MPI_INT,0,1,MPI_COMM_WORLD,&status); // blocking
    // printf("DBG.t%d: got token %d from 0\n",mpi_i,b2);
    if (!b2) // skip if error token was send
    for  (i2=0;         i2<b_ofs[pt_n]*Csize; i2+=VChunkSize) {
      if (i2+(VChunkSize)<=b_ofs[pt_n]*Csize) i3= VChunkSize;
      else                               i3=b_ofs[pt_n]*Csize-i2;
      MPI_Send(((char *)v)+i2,i3,MPI_BYTE,/*dst*/0,/*tag*/0,MPI_COMM_WORLD);
    }
  }
#endif // MPI?
#endif // MPI&0?
#if CONFIG_MPI
  { int etmp=nerr;
    MPI_Allreduce(&etmp,&nerr, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD); }
#endif
  return(nerr);
}

int loadvec(scplx *v,const char *nam,int idx,int blk){
#if CONFIG_MPI
  MPI_Status status;
#endif
  FILE *f=NULL;
  ULong n2=0;
  int i=0, nerr=0;
  char s1[128], *s2;
  if (blk>0) {
    printf("\nWarning savevec: code called from working thread, skip");
    return 0;
  }
  if (idx>=0) i=sprintf(s2=s1,nam,idx); else s2=(char *)nam;
  if (i<0 && !mpi_i) E0("fname to long");
  if ((vvv&5)==1 && !mpi_i) printf(" rv");      // short form
  if ((vvv&4)==4 && !mpi_i) printf(" r.%s",s2); // long form
  /* ToDo: do not return in a mpi-loop!, 2011-04-01 +nerr */
  /* multible open+append+close from nodes to nfs fail 2012-05-15 mkautz */
  /*   caused by scfabric errors? t=34533459 ??? nolock,rsize=1024,wsize=1024 */
  /* better send all data to thread0, which does the file IO (extra code + buffer)
    or use MPI_IO (ToDo 2012-05) */
#if CONFIG_MPI && 0 // 2012-05-15, same lock-problem as save
//  MPI_Offset my_offset;
//  MPI_File fh;
  MPI_File_open(MPI_COMM_WORLD, s2, MPI_MODE_RDONLY, MPI_INFO_NULL,&fh);
  my_offset= mpi_ofs[mpi_i]*Csize;
  MPI_File_seek(fh, my_offset, MPI_SEEK_SET);
  // MPI_File_get_position(fh, &my_current_offset);
  if (((int)b_ofs[pt_n])*Csize!=b_ofs[pt_n]*1.*Csize) F0("int_overflow");
  MPI_File_read(fh, (void *)v, b_ofs[pt_n]*Csize, MPI_CHAR, &status);
  // MPI_Get_count(&status, MPI_CHAR, &count); // ToDo: error checking
  MPI_File_close(&fh);
#else // No MPI_IO (serial + serialized MPI)
  if (mpi_i==0) {
    if (!(f=fopen(s2,"rb"))) {E1("fread %s",s2);perror("");nerr=1;} 
    /* Csize = sizeof(scplx) */
    // if ((!nerr) && fseek(f,mpi_ofs[mpi_i]*Csize,SEEK_SET)){E0("fseek");nerr=4;}
    if (v && f && (!nerr) /* && b_ofs[pt_n]*/) { // including read 0 members??
      n2 = fread((void *)v,Csize,(size_t)b_ofs[pt_n],f);
      if (n2!=b_ofs[pt_n]) {
        int n3=0;
        if (ferror(f)) {
          E2("fread1=%ld ferror=%d",(long)n2,ferror(f)); nerr=2; }
        if (!nerr)
          n3 = fread((void *)(v+n2),Csize,(size_t)b_ofs[pt_n]-n2,f);
        if (n2+n3!=b_ofs[pt_n] || ferror(f)) {
          E2("fread2=%ld ferror=%d",(long)n3,ferror(f)); nerr=2; }
      }
    } 
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
    if (mpi_n>1) { // recv from all other threads
      int i1; ULong i2, i3=0; char *buf;
      buf=(char *)malloc(VChunkSize); // 16MB   ToDo: missing error propagation!
      if (!buf) { E1("malloc%dMB", VChunkSize>>20); nerr=4; }
      if (buf) for (i1=1; i1<mpi_n; i1++)
       for (i2=mpi_ofs[i1]*Csize; i2<mpi_ofs[i1+1]*Csize; i2+=VChunkSize) {
        if (i2+(VChunkSize)        <=mpi_ofs[i1+1]*Csize) i3= VChunkSize;
        else                      i3=mpi_ofs[i1+1]*Csize-i2;
        if (f) n2 = fread(buf,1,i3,f); else n2=0; /* full node */
        if (f && n2!=i3 && nerr!=2/* only one msg */){
          E2("fread=%ld ferror=%d",(long)n2,ferror(f));nerr=2;
          sleep(60); /* give time to improve health before repeat outside */
          /* ... and reduce error logging */
        }
        /* also send on errors (no break) to avoid deadlock */
        MPI_Send(buf,i3,MPI_BYTE,/*dst*/i1,/*tag*/0,MPI_COMM_WORLD);
        /* suggest to rerun loadvec on error (temp netfs down?) */
      } // loop^2
      if (buf) free(buf);
    } // mpi_n>1
#endif
    if (f) fclose(f);
  } // mpi_i==0
#if CONFIG_MPI    // 2012-06-06 works also if locking doesnt work (NFS2,...)
  if (mpi_i!=0) { // mpi_i!=0   recv from task0
    ULong i2, i3;
    for  (i2=0;        i2< b_ofs[pt_n]*Csize; i2+=VChunkSize) {
      if (i2+(VChunkSize)<=b_ofs[pt_n]*Csize) i3= VChunkSize;
      else                               i3=(b_ofs[pt_n]*Csize)-i2;
      MPI_Recv(((char *)v)+i2, i3, MPI_BYTE,
               /*src*/0, /*tag*/0, MPI_COMM_WORLD, &status);
    }
  }
#endif // MPI?
#endif // MPI&0?
#if CONFIG_MPI
  { int etmp=nerr;
    MPI_Allreduce(&etmp,&nerr, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD); }
#endif
  return(nerr);
}

int loadvec2(scplx *v,const char *nam,int idx){ /* load v from textfile */
  FILE *f; double r; ULong j; tbase i; int k,l,ret;
  char buf[255], buf2[SZccfg];
  if (mpi_n>1) F0("missing MPI code");
  int ii=0; char s1[128],*s2;
  if(idx>=0) ii=sprintf(s2=s1,nam,idx); else s2=(char *)nam;
  if(ii<0) E0("file name to long");
  if(!v){ printf(" failed, v==NULL"); return 3;}
  if(!(f=fopen(s2,"rt"))){E1("loadvec_fopen(%s)",s2);return 1;}
  clrvec(v);ret=1;
  while(!feof(f)){
      if(!fgets(buf,255,f))break; /* zeile lesen: 0.25 |ud.3u> /n */
      if(sscanf(buf,"%lf",&r)!=1)continue;
      for(k=0;buf[k];k++) if(buf[k]=='|')break;
      i=0l;j=0l;l=2*(nn-1) CM(/2,+0,+0);
      if(buf[k]=='|'){
         for(k++;buf[k]!=0 && l>=0;k++,CM(l--,l-=2,l-=2))	/* Apr99 */
         if(buf[k]=='.'){                   } else
         if(buf[k]=='u'){i|=CM(0l,1l,1l)<<l;} else
         if(buf[k]=='d'){i|=CM(1l,2l,2l)<<l;} else
         if(buf[k]=='3'){i|=         3l <<l;} else
         E0("unknown symbol (not .ud3)"); }
      else E1("no config |xxxx> found in %s",nam);   
      j=scfg2idx(i); /* 0 .. node_n1-1 */
      if(vvv)printf("\n read: %6.3f |%s> => %ld",r,ccfg2(i,buf2),j);
      v[j]=M_d2f(r);ret=0;}
  fclose(f);
  return ret;
}

#include <stdint.h> /* uint64_t for xorshift128plus */
uint64_t xorshift_state[2]={0x65297742L,0x27651621L};
// 2018-08-20 add xorshift+ (better than mersenne-twister!? less mem=cache)
static uint64_t xorshift128plus(uint64_t s[2]) {
  uint64_t x = s[0];
  uint64_t const y = s[1];
  s[0] = y;
  x ^= x << 23; // a
  s[1] = x ^ y ^ (x >> 17) ^ (y >> 26); // b, c
  return s[1] + y;
}
            
   
#ifndef RAND_MAX
#define RAND_MAX 0x7fff
#endif
double fstime(); /* defined in sighand.c + declared in sighand.h */
/* extern int startvec; ... was removed, new: argument method */
/* randvec(): set (pseudo) random start vector, some methods available
 *  same arguments + mpi_n give same results (dbg: error reproduction)
 * called/used: by spins.c main()
 *  sum v[] during setting for faster normalization???
 *  ini could be choosen randomly to ensure different numeric errors
 *  method=startvec=0: same as 4 (but nrun unused) new: x[rnd(i)]=1
 *  method=startvec=1: x[i]=1 rest=0  (no need of normalization)  
 *     method 0 and 1 may not give full degeneracy (see doc/example_kagome36)
 *  method=startvec=2: loadvec(v,FileGS,ini)  binary-v-file
 *  method=startvec=3: loadvec2(v,FileSV,ini) text-file
 *  method=startvec=4: srand(ini) + 10bit-random every index! --default--
 *  method=startvec=5: removed, new: srand + entries (+1/-1)/sqrt(n1) only
 *  method=startvec>7: old: ini+=startvec&~7 but wrong path method>4 v=0 bug!
 *  method=startvec=8|x: rng_ini+8*nrun set outside 
 *  method=startvec>=16  rng_ini += (method&~7) since 2016-04-15
 * changes:
 *   7.12.98 JS bug!
 *   2016-04-15 add local argument method, replacing global startvec
 *     old rng_ini used via: iv+(startvec&4)*2*nruns
 *     new (iv-1)+(startvec&8)*nruns 
 * ToDo:
 *   2016-04 double fstime()? from sighand.h ?
 *   2016-04 buggy use of higher startvec-bits, missing startvec&7
 *           startvec=16 results in v=0 (bug)
 *   ToDo: first used rng_ini=iv=1, but rng_ini=iv-1=0 better
 *         x[rnd]=1 rest=0 (startvec=0)
 *         x[i]=rnd_sgn={-1,+1}/sqrt(n1)  (startvec=6?)
 *        (ToDo: later extra-norm-factor?)
 *      howto influence results? bad FTLM (highT,lowN)?
 *        n=2 start (+1,+1) or (+1,-1) ok? p.e. (1,0) missing
 *        konstruct some strange models to test trap bad random case
 *  static means: same output from same input
 * REMARK: stdlib.old: int rand; srand(int);
 * REMARK: stdlib.new: long int rand(); srandom(long int);
 * REMARK: BSD/GNU: srand(0) == srand(1) (if(seed==0)seed=1;) !!!
 *         stdlib.rand() is not thread-safe!   ToDo: multithread xorshift
 *  glibc-2.25.tar.bz2 20MB
     rand (void) ...  return (int) __random ();
     long int __random (void) ... __random_r
      int __random_r (struct random_data *buf, int32_t *result)
      PERIOD = 16*((2**31)-1)
    ToDo17: using seed(hash(myseed+mpi_idx*pt_n+pt_i) as random?
        oder index modulo 2**20 + next 2**20 random bits (need good hash)
        ggf. dummy calls on overlap
   ToDo18: local static rng_ini je nach method += 1 or = i (not outside)
      to have mode ini += 1 local, but reeinit same rvec for NEV=1
 */
void randvec(scplx *v, int method, int rng_ini){
    ULong i; double /* time */ tm1, tm2, /* norm^2 */ r2=0.0;
    rng_ini += (method>>4); /* 2016-04-15 add bit 4 and above bits */
    /* +=(method&~7) never used, buggy code! */
    method &= 7; /* 2016-04-15 fix startvec_above_7 bug */
    if (vvv) MPI_Barrier(MPI_COMM_WORLD);
#ifdef CFG_FTLM /* seen 2018-08 on sawtooth-chain J1=0.45,J2=-1 */
    if (method!=6 && !mpi_i) /* FTLM needs some level of randomness */
      printf("\nWARNING: startvec&7 != 6 may generate bad FTLM results\n");
#endif
    if (method==0 || method==4) {
      if (rng_ini!=0) { /* add test srand 2016-04-21 g++492@jPC no-stdlib */
        srand(1); i=rand(); srand(2); 
        if (rand()==(int)i && !mpi_i) printf("\nWARNING: bad srand\n");
      }
    }
    if (vvv && !mpi_i) printf("\n randvec%d-%02d", method, rng_ini);
    tm2=fstime(); /* added 2016-04-15 */
    clrvec(v); /* memory first touch for ccNUMA */
    if (method<=1 && mpi_ofs[mpi_n]){ /* set one component to 1, rest to 0 */
      /* 2016-04-15 replace v[0]=1 by v[ini%n1]=1; 04-21 correct seed=1.. ~0 */
      if (method==0) { /* added 2016-04-15 replaced method4, bad for MPI */
        srand(rng_ini+1); rng_ini = rand() % (int)mpi_ofs[mpi_n];
        MPI_Bcast(&rng_ini, 1, MPI_INT, 0, MPI_COMM_WORLD);
        if (vvv && !mpi_i) printf("-%06d",rng_ini);
      }
      if   ((ULong)rng_ini>=mpi_ofs[mpi_n  ]) rng_ini %= (int)mpi_ofs[mpi_n];
      if (  (ULong)rng_ini>=mpi_ofs[mpi_i  ]
         && (ULong)rng_ini< mpi_ofs[mpi_i+1]  )
      { v[  (ULong)rng_ini -mpi_ofs[mpi_i  ] ]=M_d2f(1); r2=1.0; }
      /* set v[0]=1 for to small n1 */
      /* if   ((ULong)rng_ini>=mpi_ofs[mpi_n  ]) v[0]=M_d2f(1); */
    } else
    if (method==2){ loadvec( v,FileGS,rng_ini,-1);   } else
    if (method==3){ loadvec2(v,FileSV,rng_ini);      } else
    if (method==4 /*|| method==0 2016-04-15 */){ /* 
       * srand(rnd);rnd++; -- bad idea, resulting in a bug until v2.17 
       * ToDo: spread in a way that norm is exactly 1.0, use l^2+m^2=n^2 
       * ToDo1104: different results for different mpi_n, better 
       *   using mypseudorandom(index+const)
       *   or randomarray[index%len]^rarray[(index/len)%len]^const
       *   or const + some random bigger values ??? */
      /* try to get near norm=1: integral(0..1) x=.5x^2 but x^2=(1/3)x^3 */
      double mean_inorm=1./(1024*sqrt(0.5*mpi_ofs[mpi_n])); /* Apr16 */
      srand(rng_ini+mpi_i+1); /* same random vector only for same mpi_n! */
      /* 2016-04-21: print first random number for debugging bad srand */
      i= rand() & ((1<<16)-1); if (vvv && !mpi_i) printf("-%04x",(unsigned)i);
      for(i=0l;i<b_ofs[pt_n];i++) {
        /* Todo2016 speedup using 3*10bits or better finite nonzero entries */
        /* Apr16: 0.000..0.999 mean=0.5 but mean(x^2)=0.333 not 0.25 ! */
        v[i]=M_d2f((1023 & rand())*mean_inorm); /* Jul00 10bits + Apr16 */
        r2 += Norm(v[i]); /* do not have exact 1.0 but near */
      }
      if (!mpi_i) { /* Todo2016: only for r2<1e-6 ? */
        r2-=Norm(v[0]);v[0]=M_d2f(0.5);r2+=0.25; /* be sure v is never zero */
      }
    } else 
    if (method==5){ /* set all v[i]=+-1 ? for FTLM, 2018=bad, use method=6 */
      // E0("startvec=5 removed in v2.35, ToDo: storeh may return i_min");
      double inorm=1./(sqrt(1.0*mpi_ofs[mpi_n])); /* Apr16 inverse norm */
      int rbits=0;  /* grep RAND_MAX /usr/include/stdlib.h # 2147483647 */
      srand(rng_ini+mpi_i+1); /* same random vector only for same mpi_n! */
      i= rand() & ((1<<16)-1); if (vvv && !mpi_i) printf("-%04x",(unsigned)i);
      for(i=0l;i<b_ofs[pt_n];i++) {
        if ((i&31)==0) rbits=rand(); /* get next 32 random bits, speedup! */
        /* Apr16: 0.000..0.999 mean=0.5 but mean(x^2)=0.333 not 0.25 ! */
        v[i]=M_d2f((1-2*(rbits&1))*inorm); /* Apr16 +-1 only */
        rbits>>=1; /* next bit, old 1bit/rand()= 0.030s, 32bits/rand()=0.008s */
        // r2 += Norm(v[i]); /* not needed, just a check of auto norm=1 */
        // if (mpi_ofs[mpi_n]-i<8) printf(" i,v[i],r2,1-r2= %8d %10.3e %10.3e %.3e\n",(int)i,D(v[i]),r2,1-r2);
        // 36+6 flt: r2= 1 + 3e-8   dbl: r2= 1 - 9e-15 
        // 34+8                     dbl: r2= 1 + 6e-12 rv.t[s]=0.030
      }
      r2=0.0; if (!mpi_i) r2=1; /* 34+8: old+ normvec=0.008s to new= 0.004s */
      // ToDo2016: check pthread-correctness!
    } else
    if (method==6 /* 2018-08-20 xorshift full random, needed for FTLM */){
      double mean_inorm=1./(((1UL<<31)*1.0*(1UL<<31))); /* get range 0..1 */
      uint64_t x1;
      srand(rng_ini+mpi_i+1); /* same random vector only for same mpi_n! */
      /* 2016-04-21: print first random number for debugging bad srand */
      i= rand() & ((1<<16)-1); if (vvv && !mpi_i) printf("-%04x",(unsigned)i);
      xorshift_state[0] ^= i; /* weak initialization? */
      for(i=0l;i<10;i++)  xorshift128plus(xorshift_state); /* shake bits */
      // printf("\nDBG: %e %016lx\n",1.*((~((uint64_t)0))>>1),((~((uint64_t)0))>>1) );
      for(i=0l;i<b_ofs[pt_n];i++) {
        x1= xorshift128plus(xorshift_state) & ((~((uint64_t)0))>>1);
        v[i]=     (1-2*(int)(x1&1))*M_d2f((x1>>1)*mean_inorm);
#if VecType & 4 /* complex vectors */
        x1= xorshift128plus(xorshift_state) & ((~((uint64_t)0))>>1);
        v[i]+=C(0,(1-2*(int)(x1&1))*M_d2f((x1>>1)*mean_inorm));
#endif
        r2 += Norm(v[i]); /* do not have exact 1.0 but near */
      }
      if (r2==0 && !mpi_i) { /* Todo2016: only for r2<1e-6 ? */
        r2-=Norm(v[0]);v[0]=M_d2f(0.5);r2+=0.25; /* be sure v is never zero */
      }
    } else { /* if (method>=7) */ 
      printf(" no_startvec (v=0)");
    } /* method? */
    // printf(" pt_n=%d b_ofs[pt_n]=%ld",pt_n,b_ofs[pt_n]);
    if (mpi_i<2 && b_ofs[pt_n]<64) { /* debugging method=6 2018-08 */
      if (!mpi_i) printf(" dbg.vrnd:\n");
      for(i=0l;i<b_ofs[pt_n] && i<4;i++)
#if VecType & 4 /* complex vectors */
         printf(" v %d.%02ld = %13.6e + %13.6e I\n", 
                                         mpi_i,i,Re(v[i]),Im(v[i]));
#else            /* reel vectors */
         printf(" v %d.%02ld = %13.6e\n",mpi_i,i,D(v[i]));
#endif           /* complex? */
    } // if (mpi_i<2 && b_ofs[pt_n]<64)
    // if (vvv) MPI_Barrier(MPI_COMM_WORLD);
    // ToDo: 
    // #ifdef CONFIG_MPI /* needed? use dummy MPI_Allreduce in block.h */
    { double tr2=r2;
      MPI_Allreduce(&tr2,&r2,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD); }
    // #endif
    if (vvv && !mpi_i) printf(".r=1%+.0e",r2-1);
    /* ToDo2016: add arg r2=norm2, if 0 autodetect, else multiply by r2^-.5 */
    if (r2 != 1.0)
     normvec(v); /* 2016-04-15 if(r2!=1) added */
    tm1=fstime()-tm2; /* added 2016-04-15 34+8: 0.008s to 0.004s */
    if (vvv && !mpi_i) printf(" rv.t[m]=%.2f",tm1/60.);
    return;
}  // randvec()
    
/* scalar product, need correct order(?): <a+ib|c+id>!=<c+id|a+ib> */
cplx sprod(scplx *a, scplx *b){ /* non-thread-parallel, tvec=scplx since v2.36 */
  cplx r=0.0; ULong i;
  if (a && b)  /* not for a2 2x2 2013-01 */
  for (i=0l;i<b_ofs[pt_n];i++) 
    r += Conj(M_f2d(a[i])) * M_f2d(b[i]);
#ifdef CONFIG_MPI
  { double add[2]={Re(r),Im(r)}, sum[2]={0,0};
    MPI_Allreduce(&add, &sum, (VecType&4)/4+1,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=C(sum[0],sum[1]); }
#endif
  return(r);
}
    
int  iortho(scplx *v0, scplx *v1){ /* 0=ok,1=parallel */
    cplx r1; double r2; ULong i;
    r1=sprod(v0,v1);
    if (Norm(r1)<1.0e-10) return(0);
    r2=0.0;
    for (i=0l;i<b_ofs[pt_n];i++){
      /* Conj(r1) for correct degeneracy for complex vectors */
      /* 2013-04-04 avoid float16 += float16 */
      v0[i]=M_d2f(M_f2d(v0[i]) - Conj(r1)*M_f2d(v1[i])); /* round error ??? */
      r2+=Norm(M_f2d(v0[i]));
    }
#ifdef CONFIG_MPI
  { double add=r2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r2=sum; }
#endif
    /* 0.995 was successfull for degeneracy of kagome-JJ N=54 u+d=51+3 q_0
     * 0.99999 results in to high degeneracy (909 instead of 72) Jul04
     * */
    if (Norm(r1)>0.995) {printf(" iortho: v0=v1");return(1);}
    r2=1.0/sqrt(r2);
//    for (i=0l;i<b_ofs[pt_n];i++)
//      v0[i]=M_d2f(M_f2d(v0[i])*r2);  /* call mltly_av ??? */
    mltply_av(r2,v0);
    return(0);
}

